from typing import Optional, Sequence

import torch
import numpy as np

from .BaseHSMM import BaseHSMM # type: ignore
from ..emission_dists import CategoricalEmissions # type: ignore


class CategoricalHSMM(BaseHSMM, CategoricalEmissions):
    """
    Categorical Hidden semi-Markov Model (HSMM)
    ----------
    Hidden semi-Markov model with categorical (discrete) emissions. This model is an extension of classical HMMs where the duration of each state is modeled by a geometric distribution.
    Duration in each state is modeled by a Categorical distribution with a fixed maximum duration.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    n_emissions (int): 
        Number of emissions in the model.
    random_state (int):
        Random seed for reproducibility.
    params_init (bool):
        Whether to initialize the model parameters prior to fitting.
    init_dist (SAMPLING_DISTRIBUTIONS):
        Distribution to use for initializing the model parameters.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial state probabilities and transition probabilities.
    verbose (bool):
        Whether to print progress logs during fitting.
    """
    def __init__(self,
                 n_states: int,
                 n_emissions: int,
                 max_duration: int,
                 alpha: float = 1.0,
                 params_init: bool = False,
                 verbose: bool = True,
                 random_state: Optional[int] = None, 
                 device: Optional[torch.device] = None):
        
        BaseHSMM.__init__(self,n_states,max_duration,params_init,alpha,verbose,random_state,device)
        
        CategoricalEmissions.__init__(self,n_states,n_emissions,params_init,alpha,random_state,device)

    def __str__(self):
        return f'CategoricalHSMM(n_states={self.n_states}, max_duration{self.max_duration}, n_emissions={self.n_emissions})'

    @property
    def params(self) -> dict:
        """Returns the parameters of the model."""
        return {
            'pi': self.initial_vector.matrix,
            'A': self.transition_matrix.matrix,
            'D': self.duration_matrix.matrix,
            'B': self.emission_matrix.matrix
        }

    @property    
    def n_fit_params(self) -> dict:
        """Return the number of trainable model parameters."""
        return {
            'states': self.n_states,
            'transitions': self.n_states**2,
            'durations': self.n_states * self.max_duration,
            'emissions': self.n_states * self.n_emissions
        }
    
    @property
    def dof(self) -> int:
        """Returns the degrees of freedom of the model."""
        return self.n_states ** 2 + self.n_states * self.n_emissions - 1
    
    def _check_theta(self, theta):
        return None
    
    def _check_sequence(self, sequence, lengths):
        if isinstance(sequence, torch.Tensor):
            assert not sequence.is_floating_point(), 'Tensor must have integer dtype'
            valid_seq = sequence.to(self.device)
        elif isinstance(sequence, np.ndarray):
            assert sequence.dtype == np.integer, 'Array must have integer dtype'
            valid_seq = torch.from_numpy(sequence).to(self.device)
        else:
            raise NotImplementedError(f'Expected torch Tensor or numpy array, got {type(sequence)}')

        if (n_dim:=valid_seq.ndim) != 1:
            raise ValueError(f'Sequence must be 1-dimensional, got shape {n_dim}')
        elif torch.max(valid_seq).item() > (self.n_emissions - 1):
            raise ValueError('Invalid emission in sequence')
        elif lengths is not None and (n:=sum(lengths)) > (T:=valid_seq.numel()):
            raise ValueError(f'Lengths sum ({n}) differ from sequence length ({T})')
        
        if lengths is not None:
            return torch.split(valid_seq, lengths)
        else:
            return [valid_seq]
        
    def map_emission(self, emission: torch.Tensor) -> torch.Tensor:
        return CategoricalEmissions.map_emission(self,emission)

    def sample_B_params(self, seed: Optional[int] = None) -> None:
        self._emission_matrix = CategoricalEmissions.sample_emissions_params(self, self.alpha, seed)

    def update_B_params(self, X: Sequence[torch.Tensor]) -> None:
        CategoricalEmissions.update_emissions_params(self,X,self._log_gamma.exp())




from typing import Optional, Sequence, Literal

import torch
import numpy as np

from .BaseHSMM import BaseHSMM # type: ignore
from ..mixtures import GaussianMixtureEmissions # type: ignore
from ..emission_dists import GaussianEmissions # type: ignore


class GaussianHSMM(BaseHSMM, GaussianEmissions):
    """
    Gaussian Hidden Semi-Markov Model (Gaussian HSMM)
    ----------
    This model assumes that the data follows a multivariate Gaussian distribution. 
    The model parameters (initial state probabilities, transition probabilities, duration probabilities,emission means, and emission covariances) 
    are learned using the Baum-Welch algorithm.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    max_duration (int):
        Maximum duration of the states.
    n_features (int):
        Number of features in the emission data.
    n_components (int):
        Number of components in the Gaussian mixture model.
    params_init (bool):
        Whether to initialize the model parameters prior to fitting.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial state probabilities and transition probabilities.
    covariance_type (COVAR_TYPES):
        Type of covariance parameters to use for the emission distributions.
    min_covar (float):
        Floor value for covariance matrices.
    random_state (Optional[int]):
        Random seed to use for reproducible results.
    verbose (bool):
        Whether to print progress logs during fitting.
    """

    COVAR_TYPES = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self,
                 n_states: int,
                 n_features: int,
                 max_duration: int,
                 params_init: bool = False,
                 alpha: float = 1.0,
                 covariance_type: COVAR_TYPES = 'full',
                 min_covar: float = 1e-3,
                 context_means: bool = False,
                 context_covars: bool = False,
                 verbose: bool = True,
                 random_state: Optional[int] = None,
                 device: Optional[torch.device] = None):

        self.context_means = context_means
        self.context_covars = context_covars

        BaseHSMM.__init__(self,n_states,max_duration,params_init,alpha,verbose,random_state,device)
        
        GaussianEmissions.__init__(self,n_states,n_features,params_init,covariance_type,
                                   min_covar,random_state,device)
                    
    def __str__(self):
        return f'GaussianHSMM(n_states={self.n_states}, n_durations={self.max_duration},n_features={self.n_features})'

    @property
    def n_fit_params(self):
        """Return the number of trainable model parameters."""
        return {
            'states': self.n_states,
            'transitions': self.n_states**2,
            'durations': self.n_states * self.max_duration,
            'means': self.n_states * self.n_features,
            'covars': {
                'spherical': self.n_states,
                'diag': self.n_states * self.n_features,
                'full': self.n_states * self.n_features * (self.n_features + 1) // 2,
                'tied': self.n_features * (self.n_features + 1) // 2,
            }[self.covariance_type],
        }
    
    @property
    def params(self):
        return {'pi': self.initial_vector.matrix,
                'A': self.transition_matrix.matrix,
                'D': self.duration_matrix.matrix,
                'B':{'means': self.means, 
                        'covars': self.covs}}

    @property
    def dof(self):
        return self.n_states**2 - 1 + self.means.numel() + self.covs.numel() 
    
    def _check_sequence(self, sequence, lengths=None) -> Sequence[torch.Tensor]:
        if isinstance(sequence, torch.Tensor):
            valid_seq = sequence.to(device=self.device)
        elif isinstance(sequence, np.ndarray):
            valid_seq = torch.from_numpy(sequence).to(dtype=torch.int64, device=self.device)
        else:
            raise NotImplementedError(f'Expected torch Tensor or numpy array, got {type(sequence)}')
        
        if valid_seq.ndim != 2:
            raise ValueError(f'Sequence must have shape (T,{self.n_features}). Got {sequence.shape}.')
        elif valid_seq.shape[1] != self.n_features:
            raise ValueError(f'Sequence must have shape (T,{self.n_features}). Got {sequence.shape}.')
        elif lengths is not None and sum(lengths) > valid_seq.size(dim=0):
            raise ValueError(f'Lengths sum ({sum(lengths)}) differ from sequence length ({valid_seq.size(dim=0)})')
       
        if lengths is not None:
            return torch.split(valid_seq, lengths)
        else:
            return [valid_seq]
        
    def _check_theta(self, theta: Optional[torch.Tensor]=None) -> Optional[torch.Tensor]:
        """Check the given contextual variables."""
        if (self.context_covars or self.context_means) and theta is None:
            raise ValueError('Theta must be given when using context means or covars.')
        elif theta is not None and not (self.context_covars or self.context_means):
            raise ValueError('Theta must be None when NOT using context means or covars.')
        else:
            if theta is None:
                return theta
            else:
                return torch.hstack((theta, torch.ones(size=(theta.shape[0],1), 
                                                       dtype=torch.float64, 
                                                       device=self.device)))
    
    def map_emission(self, emission: torch.Tensor) -> torch.Tensor:
        return GaussianEmissions.map_emission(self,emission)

    def sample_B_params(self, seed: Optional[int] = None) -> None:
        self._means,self._covs = GaussianEmissions.sample_emissions_params(self, seed)

    def update_B_params(self, X: Sequence[torch.Tensor]) -> None:
        GaussianEmissions.update_emissions_params(self,X,self._log_gamma.exp())


class GaussianMixtureHSMM(BaseHSMM, GaussianMixtureEmissions):
    """
    Gaussian Hidden Semi-Markov Model (Gaussian HSMM)
    ----------
    This model assumes that the data follow a multivariate Gaussian distribution. 
    The model parameters (initial state probabilities, transition probabilities, duration probabilities,emission means, and emission covariances) 
    are learned using the Baum-Welch algorithm.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    max_duration (int):
        Maximum duration of the states.
    n_features (int):
        Number of features in the emission data.
    n_components (int):
        Number of components in the Gaussian mixture model.
    params_init (bool):
        Whether to initialize the model parameters prior to fitting.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial state probabilities and transition probabilities.
    covariance_type (COVAR_TYPES):
        Type of covariance parameters to use for the emission distributions.
    min_covar (float):
        Floor value for covariance matrices.
    random_state (Optional[int]):
        Random seed to use for reproducible results.
    verbose (bool):
        Whether to print progress logs during fitting.
    """

    COVAR_TYPES = Literal['spherical', 'tied', 'diag', 'full']

    def __init__(self,
                 n_states: int,
                 n_features: int,
                 max_duration: int,
                 n_components: int = 1,
                 params_init: bool = False,
                 alpha: float = 1.0,
                 covariance_type: COVAR_TYPES = 'full',
                 min_covar: float = 1e-3,
                 context_means: bool = False,
                 context_covars: bool = False,
                 verbose: bool = True,
                 random_state: Optional[int] = None,
                 device: Optional[torch.device] = None):

        self.context_means = context_means
        self.context_covars = context_covars

        BaseHSMM.__init__(self,n_states,max_duration,params_init,alpha,verbose,random_state,device)
        
        GaussianMixtureEmissions.__init__(self,n_states,n_components,n_features,params_init,alpha,covariance_type,
                                          min_covar,random_state,device)
                    
    def __str__(self):
        return f'GaussianMixtureHSMM(n_states={self.n_states}, n_durations={self.max_duration},n_features={self.n_features}, n_components={self.n_components})'

    @property
    def n_fit_params(self):
        """Return the number of trainable model parameters."""
        fit_params_dict = {
            'states': self.n_states,
            'transitions': self.n_states**2,
            'durations': self.n_states * self.max_duration,
            'weights': self.n_states * self.n_components,
            'means': self.n_states * self.n_features * self.n_components,
            'covars': {
                'spherical': self.n_states,
                'diag': self.n_states * self.n_features,
                'full': self.n_states * self.n_features * (self.n_features + 1) // 2,
                'tied': self.n_features * (self.n_features + 1) // 2,
            }[self.covariance_type],
        }

        return fit_params_dict
    
    @property
    def params(self):
        param_dict = {'pi': self.initial_vector.matrix,
                    'A': self.transition_matrix.matrix,
                    'D': self.duration_matrix.matrix,
                    'B':{'weights': self.weights.matrix,
                         'means': self.means, 
                         'covars': self.covs}}
        return param_dict

    @property
    def dof(self):
        return self.n_states**2 - 1 + self.n_states*self.n_components - self.n_states + self.means.numel() + self.covs.numel() 
    
    def _check_sequence(self, sequence, lengths=None) -> Sequence[torch.Tensor]:
        if isinstance(sequence, torch.Tensor):
            valid_seq = sequence.to(self.device)
        elif isinstance(sequence, np.ndarray):
            valid_seq = torch.from_numpy(sequence).to(self.device)
        else:
            raise NotImplementedError(f'Expected torch Tensor or numpy array, got {type(sequence)}')
        
        if (n_dim:=valid_seq.ndim) != 2:
            raise ValueError(f'Sequence must be 2-dimensional. Got {n_dim}.')
        elif (F:=valid_seq.shape[1]) != self.n_features:
            raise ValueError(f'Sequence 2nd dim must be {self.n_features}. Got {F}.')
        elif lengths is not None and (n:=sum(lengths)) > (T:=valid_seq.size(dim=0)):
            raise ValueError(f'Lengths sum ({n}) differ from sequence length ({T})')
       
        if lengths is not None:
            return torch.split(valid_seq, lengths)
        else:
            return [valid_seq]
        
    def _check_theta(self, theta: Optional[torch.Tensor]=None) -> Optional[torch.Tensor]:
        """Check the given contextual variables."""
        if (self.context_covars or self.context_means) and theta is None:
            raise ValueError('Theta must be given when using context means or covars.')
        elif theta is not None and not (self.context_covars or self.context_means):
            raise ValueError('Theta must be None when NOT using context means or covars.')
        else:
            if theta is None:
                return theta
            else:
                return torch.hstack((theta, torch.ones(size=(theta.shape[0],1), 
                                                       dtype=torch.float64, 
                                                       device=self.device)))
    
    def map_emission(self, emission: torch.Tensor) -> torch.Tensor:
        return GaussianMixtureEmissions.map_emission(self,emission)

    def sample_B_params(self, seed: Optional[int] = None) -> None:
        self.means = GaussianMixtureEmissions.sample_emissions_params(self, seed)

    def update_B_params(self, X: Sequence[torch.Tensor]) -> None:
        posterior = torch.zeros(size=(self.n_sequences,self.n_states,self.n_components,self.n_samples), 
                                dtype=torch.float64, 
                                device=self.device)
        
        for i, seq in enumerate(X):
            posterior[i] = torch.exp(self._log_gamma[i,:,:seq.size(0)].unsqueeze(dim=-2).expand(self.n_states,self.n_components,-1) + self.compute_responsibilities(seq))

        GaussianMixtureEmissions.update_emissions_params(self,X,posterior)


import torch
from typing import Sequence, Tuple, Optional
from prettytable import PrettyTable

class ConvergenceHandler:
    """
    Convergence Monitor
    ----------
    Convergence monitor for HMM training. Stores the score at each iteration and checks for convergence.

    Parameters
    ----------
    max_iter : int
        Maximum number of iterations.
    n_init : int
        Number of initializations.
    tol : float
        Convergence threshold.
    post_conv_iter : int
        Number of iterations to run after convergence.
    verbose : bool
        Print convergence information.
    """

    def __init__(self, 
                 max_iter: int,
                 n_init:int , 
                 tol: float, 
                 post_conv_iter: int,
                 device: torch.device, 
                 verbose: bool = True):
        
        self.tol: float = tol
        self.verbose: bool = verbose
        self.post_conv_iter: int = post_conv_iter
        self.device: torch.device = device
        self.iter: int = 0
        self.score = torch.full(size=(max_iter, n_init),
                                fill_value=float('nan'), 
                                dtype=torch.float64,
                                device=self.device)
                                
        self.delta = self.score.clone()
        
    def _update(self, new_score: float, rank: int):
        """Update the iteration count."""
        old_score = self.score[self.iter-1, rank]
        self.score[self.iter, rank] = new_score
        self.delta[self.iter, rank] = new_score - old_score

    def check(self, new_score: float, rank: int) -> bool:
        """Check if the model has converged and update the convergence monitor."""    
        # Store prev score and update current score
        self._update(new_score, rank)

        # Update iteration count and check convergence
        if self.iter < self.post_conv_iter:
            is_converged = False
        else:
            is_converged = bool(torch.all(self.delta[(self.iter-self.post_conv_iter):self.iter, rank] < self.tol).item())

        # Print convergence information
        if self.verbose:
            print(f"""Iteration: {self.iter+1} | Score: {new_score:.2f} | Delta: {self.delta[self.iter, rank].item():.2f} | Converged = {is_converged}""")            

        return is_converged

def states_names(n: int, state_type: str) -> Sequence[str]:
    type_alias = state_type[0].upper()
    return [f'{type_alias}_{n}' for n in range(n)]

def laplace_smoothing(log_matrix: torch.Tensor, k: float) -> torch.Tensor:
    """Laplace smoothing of a log probability matrix"""

    if k <= 0.0:  
        return log_matrix
    else:
        real_matrix = torch.exp(log_matrix) 
        smoothed_log_matrix = torch.log((real_matrix + k) / (1 + k * real_matrix.shape[-1]))
        return smoothed_log_matrix
    
def init_matrix(prior: float, target_size: Tuple[int], semi: bool = False) -> torch.Tensor:
    
    alphas = torch.full(size=target_size, 
                        fill_value=prior, 
                        dtype=torch.float64)
    
    probs = torch.distributions.Dirichlet(alphas).sample()

    if semi:
        probs.fill_diagonal_(0)
        probs /= probs.sum(dim=-1, keepdim=True)

    return probs

def validate_means(means: torch.Tensor, n_states: int, n_features: int, n_components: Optional[int]=None) -> torch.Tensor:
    """Do basic checks on matrix mean sizes and values"""
    if n_components is None:
        valid_shape = (n_states, n_features)
        if (n_dim:=means.ndim) != 2:
            raise ValueError(f"Tensor must be 2D, got {n_dim} dims instead")
        elif (m_shape:=means.shape) != valid_shape:
            raise ValueError(f"Tensor must have shape {valid_shape}, got {m_shape} instead")
    else:
        valid_shape = (n_states, n_components, n_features)
        if (n_dim:=means.ndim) != 3:
            raise ValueError(f"Tensor must be 3D, got {n_dim} dims instead")
        elif (m_shape:=means.shape) != valid_shape:
            raise ValueError(f"Tensor must have shape {valid_shape}, got {m_shape} instead")

    if torch.any(torch.isnan(means)):
        raise ValueError("means must not contain NaNs")
    elif torch.any(torch.isinf(means)):
        raise ValueError("means must not contain infinities")
    else:
        return means
    
def validate_lambdas(lambdas: torch.Tensor, n_states: int, n_features: int) -> torch.Tensor:
    """Do basic checks on matrix mean sizes and values"""
    
    if len(lambdas.shape) != 2:
        raise ValueError("lambdas must have shape (n_states, n_features)")
    elif lambdas.shape[0] != n_states:
        raise ValueError("lambdas must have shape (n_states, n_features)")
    elif lambdas.shape[1] != n_features:
        raise ValueError("lambdas must have shape (n_states, n_features)")
    elif torch.any(torch.isnan(lambdas)):
        raise ValueError("lambdas must not contain NaNs")
    elif torch.any(torch.isinf(lambdas)):
        raise ValueError("lambdas must not contain infinities")
    elif torch.any(lambdas <= 0):
        raise ValueError("lambdas must be positive")
    else:
        return lambdas

def validate_covars(covars: torch.Tensor, 
                    covariance_type: str, 
                    n_states: int, 
                    n_features: int,
                    n_components: Optional[int]=None) -> torch.Tensor:
    """Do basic checks on matrix covariance sizes and values"""
    if n_components is None:
        valid_shape = torch.Size((n_states, n_features, n_features))
    else:
        valid_shape = torch.Size((n_states, n_components, n_features, n_features))    

    if covariance_type == 'spherical':
        if len(covars) != n_features:
            raise ValueError("'spherical' covars have length n_features")
        elif torch.any(covars <= 0): 
            raise ValueError("'spherical' covars must be positive")
    elif covariance_type == 'tied':
        if covars.shape[0] != covars.shape[1]:
            raise ValueError("'tied' covars must have shape (n_dim, n_dim)")
        elif (not torch.allclose(covars, covars.T) or torch.any(covars.symeig(eigenvectors=False).eigenvalues <= 0)):
            raise ValueError("'tied' covars must be symmetric, positive-definite")
    elif covariance_type == 'diag':
        if len(covars.shape) != 2:
            raise ValueError("'diag' covars must have shape (n_features, n_dim)")
        elif torch.any(covars <= 0):
            raise ValueError("'diag' covars must be positive")
    elif covariance_type == 'full':
        if len(covars.shape) != 3:
            raise ValueError("'full' covars must have shape (n_features, n_dim, n_dim)")
        elif covars.shape[1] != covars.shape[2]:
            raise ValueError("'full' covars must have shape (n_features, n_dim, n_dim)")
        for n, cv in enumerate(covars):
            eig_vals, _ = torch.linalg.eigh(cv)
            if (not torch.allclose(cv, cv.T) or torch.any(eig_vals <= 0)):
                raise ValueError(f"component {n} of 'full' covars must be symmetric, positive-definite")
    else:
        raise NotImplementedError(f"This covariance type is not implemented: {covariance_type}")
    
    return covars
       
def init_covars(tied_cv: torch.Tensor, 
                covariance_type: str, 
                n_states: int) -> torch.Tensor:
    """Initialize covars to a given covariance type"""

    if covariance_type == 'spherical':
        return tied_cv.mean() * torch.ones((n_states,))
    elif covariance_type == 'tied':
        return tied_cv
    elif covariance_type == 'diag':
        return tied_cv.diag().unsqueeze(0).expand(n_states, -1)
    elif covariance_type == 'full':
        return tied_cv.unsqueeze(0).expand(n_states, -1, -1)
    else:
        raise NotImplementedError(f"This covariance type is not implemented: {covariance_type}")
    
def fill_covars(covars: torch.Tensor, 
                covariance_type: str, 
                n_states: int, 
                n_features: int,
                n_components: Optional[int]=None) -> torch.Tensor:
    """Fill in missing values for covars"""
    
    if covariance_type == 'full':
        return covars
    elif covariance_type == 'diag':
        return torch.stack([torch.diag(covar) for covar in covars])
    elif covariance_type == 'tied':
        return covars.unsqueeze(0).expand(n_states, -1, -1)
    elif covariance_type == 'spherical':
        eye = torch.eye(n_features).unsqueeze(0)
        return eye * covars.unsqueeze(-1).unsqueeze(-1)
    else:
        raise NotImplementedError(f"This covariance type is not implemented: {covariance_type}")

def print_table(rows: list, header: list, title: str):
    """
    Helper method for the pretty print function. It prints the parameters
    as a nice table.
    """
    t = PrettyTable(title=title, 
                    field_names=header, 
                    header_style='upper',
                    padding_width=1, 
                    title_style='upper')
    
    for row in rows:
        t.add_row(row)
    
    print(t)


from typing import Optional, Sequence, Tuple, Dict, List, Union
from abc import ABC, abstractmethod, abstractproperty

import torch
import numpy as np # type: ignore
import matplotlib.pyplot as plt # type: ignore

from ..stochastic_matrix import TransitionMatrix, ProbabilityVector # type: ignore
from ..utils import ConvergenceHandler # type: ignore


class BaseHMM(ABC): 
    """
    Base Abstract Class for HMM
    ----------
    Base Class of Hidden Markov Models (HMM) class that provides a foundation for building specific HMM models.
    """

    def __init__(self,
                 n_states: int,
                 params_init: bool = False,
                 alpha: float = 1.0,
                 verbose: bool = True,
                 random_state: Optional[int] = None,
                 device: Optional[torch.device] = None):

        self.n_states = n_states
        self.alpha = alpha
        self.seed = random_state
        self.verbose = verbose

        if device is None:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = device

        if params_init:
            self._initial_vector, self._transition_matrix = self.sample_chain_params(alpha,random_state)

    @property
    def transition_matrix(self) -> TransitionMatrix:
        return self._transition_matrix

    @transition_matrix.setter
    def transition_matrix(self, matrix):
        assert (self.n_states,self.n_states) == matrix.shape, 'Matrix dimensions differ from HMM model'
        if isinstance(matrix, TransitionMatrix):
            self._transition_matrix = matrix
        elif isinstance(matrix, torch.Tensor):
            self._transition_matrix = TransitionMatrix(n_states=self.n_states,
                                                       matrix=matrix,
                                                       device=self.device)
        else:
            raise NotImplementedError('Matrix type not supported')

    @property
    def initial_vector(self) -> ProbabilityVector:
        return self._initial_vector

    @initial_vector.setter
    def initial_vector(self, vector):
        assert (self.n_states,) == vector.shape, 'Matrix dimensions differ from HMM model'
        if isinstance(vector, ProbabilityVector):
            self._initial_vector = vector
        elif isinstance(vector, torch.Tensor):
            self._initial_vector = ProbabilityVector(n_states=self.n_states, 
                                                     vector=vector,
                                                     device=self.device)
        else:
            raise NotImplementedError('Matrix type not supported')

    @property
    def fit_score(self) -> float:
        if len(self.train_lengths) == 1:
            return self._log_alpha[0,:,-1].logsumexp(dim=0).item()
        else:
            score = 0.0
            for i, length in enumerate(self.train_lengths):
                score += self._log_alpha[i,:,length-1].logsumexp(dim=0).item()
            return score
        
    @property
    def _check_params(self):
        """Check if the model parameters are set."""
        return self.params

    @abstractproperty
    def params(self) -> Dict[str, torch.Tensor]:
        """Returns the parameters of the model."""
        pass

    @abstractproperty   
    def n_fit_params(self) -> Dict[str, int]:
        """Return the number of trainable model parameters."""
        pass
    
    @abstractproperty
    def dof(self) -> int:
        """Returns the degrees of freedom of the model."""
        pass

    @abstractmethod
    def map_emission(self, emission: torch.Tensor) -> torch.Tensor:
        """Get emission probabilities for a given sequence of observations."""
        pass

    @abstractmethod
    def _check_sequence(self, 
                        sequence: Union[torch.Tensor,np.ndarray], 
                        lengths: Optional[List[int]]=None) -> Sequence[torch.Tensor]:
        """Check if the sequence is valid, encode, transform if necessary."""
        pass

    @abstractmethod
    def _check_theta(self, theta: Optional[torch.Tensor]=None) -> Optional[torch.Tensor]:
        """Returns the parameters of the model."""
        pass

    @abstractmethod
    def sample_B_params(self, seed: Optional[int] = None):
        """Sample the emission parameters."""
        pass

    @abstractmethod
    def update_B_params(self, X: Sequence[torch.Tensor]):
        """Update the emission parameters."""
        pass

    def sample_chain_params(self, alpha: float, seed: Optional[int] = None) -> Tuple[ProbabilityVector, TransitionMatrix]:
        """Initialize the model parameters."""
        return (ProbabilityVector(self.n_states, rand_seed=seed, alpha=alpha, device=self.device), 
                TransitionMatrix(self.n_states, rand_seed=seed, alpha=alpha, device=self.device))

    def _init_train_vars(self,
                         sequence_list: Sequence[torch.Tensor],
                         lengths: Optional[Sequence[int]]=None,
                         theta: Optional[torch.Tensor]=None):
        """Initialize the variables for the forward-backward algorithm."""
        self.n_sequences = len(sequence_list)
        self.n_samples = sequence_list[0].size(0) if lengths is None else max(lengths)
        self.train_lengths = [self.n_samples] if lengths is None else lengths
        
        self._log_alpha = torch.zeros(size=(self.n_sequences, self.n_states, self.n_samples), 
                                      dtype=torch.float64,
                                      device=self.device)
        self._log_beta = torch.zeros(size=(self.n_sequences, self.n_states, self.n_samples), 
                                     dtype=torch.float64, 
                                     device=self.device)
        self._log_gamma = torch.zeros(size=(self.n_sequences, self.n_states, self.n_samples), 
                                      dtype=torch.float64, 
                                      device=self.device)
        self._log_xi = torch.zeros(size=(self.n_sequences, self.n_samples - 1, self.n_states, self.n_states),
                                   dtype=torch.float64, 
                                   device=self.device)

    def _reset_train_vars(self):
        """Reset the variables for the forward-backward algorithm."""
        self._log_alpha.zero_()
        self._log_beta.zero_()
        self._log_gamma.zero_()
        self._log_xi.zero_()
        
    def _viterbi(self, X:torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Viterbi algorithm for decoding the most likely sequence of hidden states."""
        # Check if params are initialized
        self._check_params

        # Tranform sequence to indices
        X_valid = self._check_sequence(X)[0]
        n_samples = X_valid.shape[0]

        # Initialize matrices
        viterbi_path = torch.empty(size=(n_samples,), 
                                   dtype=torch.int32,
                                   device=self.device)
        
        viterbi_prob = torch.empty(size=(self.n_states, n_samples), 
                                       dtype=torch.float64,
                                       device=self.device)
        psi = viterbi_prob.clone()

        # Initialize first column
        viterbi_prob[:,0] = self._initial_vector + self.map_emission(X_valid[0])
        for t in range(1, n_samples):
            trans_seq = viterbi_prob[:,t-1] + self.map_emission(X_valid[t])
            trans_seq = self._transition_matrix + trans_seq.reshape((-1, 1))
            viterbi_prob[:,t] = torch.max(trans_seq, dim=0).values
            psi[:,t] = torch.argmax(trans_seq, dim=0)

        # Backtrack the most likely sequence
        viterbi_path[-1] = torch.argmax(viterbi_prob[:, -1])
        for t in reversed(range(n_samples - 1)):
            viterbi_path[t] = psi[viterbi_path[t + 1], t + 1]

        return viterbi_path, viterbi_prob

    def _forward(self, X:Sequence[torch.Tensor]):
        """Forward pass of the forward-backward algorithm."""
        for i,seq in enumerate(X):
            self._log_alpha[i,:,0].add_(self._initial_vector + self.map_emission(seq[0]))
            for t in range(1, self.train_lengths[i]):
                self._log_alpha[i,:, t].add_(
                    self.map_emission(seq[t])
                    + torch.logsumexp(self._log_alpha[i,:, t - 1].reshape(-1, 1) + self._transition_matrix, dim=0)
                )

    def _backward(self, X:Sequence[torch.Tensor]):
        """Backward pass of the forward-backward algorithm."""
        for i,seq in enumerate(X):
            for t in reversed(range(self.train_lengths[i] - 1)):
                self._log_beta[i,:,t].add_(torch.logsumexp(
                    self._transition_matrix
                    + self.map_emission(seq[t+1])
                    + self._log_beta[i,:, t + 1], dim=1)
                )

    def _gamma(self):
        """Compute the log-Gamma variable in Hidden Markov Model."""
        denom = self._log_alpha + self._log_beta
        self._log_gamma.add_(denom - torch.logsumexp(denom, dim=1, keepdim=True))

    def _xi(self, X:Sequence[torch.Tensor]):
        """Compute the log-Xi variable in Hidden Markov Model."""
        for i,seq in enumerate(X):
            for t in range(self.train_lengths[i] - 1):
                self._log_xi[i,t,:,:].add_(
                    self._log_alpha[i,:, t].reshape(-1, 1)
                    + self._transition_matrix
                    + self.map_emission(seq[t+1])
                    + self._log_beta[i,:,t+1]
                )
                max_t = self.train_lengths[i]
                self._log_xi[i,t,:,:].sub_(torch.logsumexp(self._log_alpha[i,:,max_t-1], dim=0))

    def _E_step(self, X: Sequence[torch.Tensor]):
        """Execute the forward-backward algorithm and compute the log-Gamma and log-Xi variables."""
        self._forward(X)
        self._backward(X)
        self._gamma()
        self._xi(X)
    
    def _M_step(self, X, theta) -> Dict[str, torch.Tensor]:
        """Compute the updated parameters for the model."""

        # Compute the posteriors
        self._E_step(X)

        return {
            'pi': self._accum_pi(),
            'A': self._accum_A()
        }
    
    def _accum_pi(self) -> torch.Tensor:
        """Accumulate the statistics for the initial vector."""
        seq_sum = self._log_gamma[:,:,0].logsumexp(dim=0)
        normed_sum = seq_sum - seq_sum.logsumexp(dim=0)
        return normed_sum

    def _accum_A(self) -> torch.Tensor:
        """Accumulate the statistics for the transition matrix."""
        seq_sum = torch.logsumexp(self._log_xi, dim=(0,1))
        normed_sum = seq_sum - seq_sum.logsumexp(dim=1, keepdim=True)
        return normed_sum
    
    def _update_chain(self, X: Sequence[torch.Tensor], theta: Optional[torch.Tensor] = None):
        """Update the model parameters."""
        new_params = self._M_step(X, theta)
        self._transition_matrix.matrix.copy_(new_params['A'])
        self._initial_vector.matrix.copy_(new_params['pi'])

    def fit(self,
            X: torch.Tensor,
            tol: float = 1e-2,
            max_iter: int = 20,
            n_init: int = 1,
            alpha: float = 1.0,
            post_conv_iter: int = 3,
            plot_score: bool = False,
            ignore_conv: bool = False,
            lengths: Optional[List[int]] = None,
            theta: Optional[torch.Tensor] = None) -> Tuple[Dict[str, torch.Tensor], torch.Tensor]:
        """Fit the model to the given sequence using the EM algorithm."""
        
        # Check if params are initialized
        self._check_params

        # Check sequence parameters and transform if lengths
        X_valid = self._check_sequence(X, lengths)
        valid_theta = self._check_theta(theta)
        self._init_train_vars(X_valid, lengths, valid_theta)

        # Initialize convergence handler
        conv = ConvergenceHandler(tol=tol, 
                                  max_iter=max_iter, 
                                  n_init=n_init,
                                  post_conv_iter=post_conv_iter,
                                  device=self.device, 
                                  verbose=self.verbose)

        distinct_models = {}
        for i in range(n_init):
            if i > 0:
                conv.iter = 0
                self.sample_chain_params(alpha)
                self.sample_B_params()
            
            for _ in range(max_iter):
                # Reset training variables
                self._reset_train_vars()
            
                # Estimate and Update Model parameters
                self._update_chain(X_valid, theta)
                self.update_B_params(X_valid)

                # Check convergence
                if conv.check(new_score=self.fit_score, rank=i) and not ignore_conv:
                    break
                
                # Update iteration count and reset training variables
                conv.iter += 1

            # Save model parameters
            distinct_models[i] = self.params

        # Update params with best model based on score (log-likelihood)
        masked_conv = conv.score.masked_fill(torch.isnan(conv.score), float('-inf'))
        max_scores = torch.max(masked_conv, dim=0).values
        best_model = distinct_models[int(torch.argmax(max_scores).item())]

        # TODO: Set model parameters to best model
        # self._set_params(best_model)

        if plot_score:
            # Define input for plot
            labels = [f'Log-likelihood - Run #{i+1}' for i in range(n_init)]

            # Plot setting
            plt.style.use('ggplot')
            _, ax = plt.subplots(figsize=(10, 7))
            ax.plot(torch.arange(max_iter), 
                    conv.score.cpu(), 
                    linewidth=2, 
                    marker='o', 
                    markersize=5, 
                    label=labels)
            
            ax.set_title('HMM Model Log-Likelihood Convergence')
            ax.set_xlabel('# Iterations')
            ax.set_ylabel('Log-likelihood')
            ax.legend(loc='lower right')
            plt.show()
        
        return best_model, max_scores

    def predict(self, X: torch.Tensor) -> torch.Tensor:
        """Compute the most likely sequence of hidden states."""
        viterbi_path, _ = self._viterbi(X)
        return viterbi_path

    def predict_proba(self, X: torch.Tensor) -> torch.Tensor:
        """Compute the log probability of the most likely sequence of hidden states."""
        _, viterbi_prob = self._viterbi(X)
        return viterbi_prob 
    
    def score(self, X: torch.Tensor, lengths: Optional[Sequence[int]] = None) -> float:
        """Compute the log-likelihood of the given sequence."""
        X_valid = self._check_sequence(X)
        self._init_train_vars(X_valid,lengths)
        self._forward(X_valid)
        return self.fit_score

    def ic(self, X: torch.Tensor, criterion: str = 'AIC') -> float:
        """
        Calculates the information criteria for a given model.

        Parameters:
        ----------
        criterion : str
            The information criterion to compute. Options are 'AIC', 'BIC', and 'HQC'. Default is 'AIC'.
        """

        dof = self.dof
        log_likelihood = self.score(X)
        
        if criterion == 'BIC':
            return -2.0 * log_likelihood + dof * np.log(X.shape[0])
        elif criterion == 'AIC':
            return -2.0 * log_likelihood + 2.0 * dof
        elif criterion == 'HQC':
            return -2.0 * log_likelihood + 2.0 * dof * np.log(np.log(X.shape[0]))
        else:
            raise NotImplementedError(f'{criterion} is not a valid information criterion. Valid options are "AIC", "BIC", and "HQC".')

    def view_params(self):
        """Print the model parameters."""
        for param in self.params.values():
            param.view()
            print('\n')








from typing import Optional, Sequence

import torch
import numpy as np

from .BaseHMM import BaseHMM # type: ignore
from ..emission_dists import CategoricalEmissions # type: ignore

class CategoricalHMM(BaseHMM, CategoricalEmissions):
    """
    Categorical Hidden Markov Model (HMM)
    ----------
    Hidden Markov model with categorical (discrete) emissions. This model is a special case of the HSMM model with a geometric duration distribution.

    Parameters:
    ----------
    n_states (int):
        Number of hidden states in the model.
    n_emissions (int): 
        Number of emissions in the model.
    random_state (int):
        Random seed for reproducibility.
    params_init (bool):
        Whether to initialize the model parameters.
    alpha (float):
        Dirichlet concentration parameter for the prior over initial distribution, transition amd emission probabilities.
    verbose (bool):
        Whether to print progress logs during fitting.
    device (torch.device):
        Device on which to fit the model.
    """

    def __init__(self,
                 n_states: int,
                 n_emissions: int,
                 alpha: float = 1.0,
                 params_init: bool = False,
                 verbose: bool = True,
                 random_state: Optional[int] = None, 
                 device: Optional[torch.device] = None):
        
        BaseHMM.__init__(self,n_states,params_init,alpha,verbose,random_state,device)
        
        CategoricalEmissions.__init__(self, n_states, n_emissions,params_init,alpha,random_state,device)
        
    def __str__(self):
        return f'CategoricalHMM(n_states={self.n_states}, n_emissions={self.n_emissions})'

    @property
    def params(self):
        return {
            'pi': self.initial_vector.matrix,
            'A': self.transition_matrix.matrix,
            'B': self.emission_matrix.matrix
        }

    @property    
    def n_fit_params(self):
        return {
            'initial_states': self.n_states,
            'transitions': self.n_states**2,
            'emissions': self.n_states * self.n_emissions
        }

    @property
    def dof(self):
        return self.n_states ** 2 + self.n_states * self.n_emissions - self.n_states - 1
    
    def _check_theta(self, theta):
        return None

    def _check_sequence(self, sequence, lengths):
        if isinstance(sequence, torch.Tensor):
            assert not sequence.is_floating_point(), 'Tensor must have integer dtype'
            valid_seq = sequence.to(self.device)
        elif isinstance(sequence, np.ndarray):
            assert sequence.dtype == np.integer, 'Array must have integer dtype'
            valid_seq = torch.from_numpy(sequence).to(self.device)
        else:
            raise NotImplementedError(f'Expected torch Tensor or numpy array, got {type(sequence)}')

        if (n_dim:=valid_seq.ndim) != 1:
            raise ValueError(f'Sequence must be 1-dimensional, got shape {n_dim}')
        elif torch.max(valid_seq).item() > (self.n_emissions - 1):
            raise ValueError('Invalid emission in sequence')
        elif lengths is not None and (n:=sum(lengths)) > (T:=valid_seq.numel()):
            raise ValueError(f'Lengths sum ({n}) differ from sequence length ({T})')
        
        if lengths is not None:
            return torch.split(valid_seq, lengths)
        else:
            return [valid_seq]

    def map_emission(self, emission: torch.Tensor) -> torch.Tensor:
        return CategoricalEmissions.map_emission(self,emission)

    def sample_B_params(self, seed: Optional[int] = None) -> None:
        self._emission_matrix = CategoricalEmissions.sample_emissions_params(self, self.alpha, seed)

    def update_B_params(self, X: Sequence[torch.Tensor]) -> None:
        CategoricalEmissions.update_emissions_params(self,X,self._log_gamma.exp())




from setuptools import setup #type: ignore

setup(
    name='chadhmm',
    packages = ['chadhmm','chadhmm/hmm','chadhmm/hsmm'],
    version='0.2.2',
    description='Package for Hidden (Semi-)Markov Models',
    author='GarroshIcecream',
    author_email='ad.pesek13@gmail.com',
    url='https://github.com/GarroshIcecream/ChadHMM',
    download_url = 'https://github.com/GarroshIcecream/ChadHMM/archive/refs/tags/v0.2.tar.gz',
    license='MIT',
    keywords=['Hidden Markov Models','Hidden Semi-Markov Models','hsmm','hmm','Gaussian Mixture Models', 'gmm'],
    install_requires=['torch','matplotlib','scikit-learn','numpy','prettytable'],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
    ]
)